"""Command Line Interface.

This module contains the command line interface (CLI) for MicroStructPy.
The CLI primarily reads XML input files and creates a microstructure according
to those inputs. It can also run demo input files.

"""


from __future__ import division
from __future__ import print_function

import argparse
import collections
import glob
import os
import shutil
import subprocess

import numpy as np
import scipy.stats
import xmltodict
from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

from microstructpy import _misc
from microstructpy import geometry
from microstructpy import seeding
from microstructpy import verification
from microstructpy.meshing import PolyMesh
from microstructpy.meshing import TriMesh
from microstructpy.meshing.trimesh import facet_check

__author__ = 'Kenneth (Kip) Hart'


# --------------------------------------------------------------------------- #
#                                                                             #
# Main Function                                                               #
#                                                                             #
# --------------------------------------------------------------------------- #
def main():
    """CLI calling function"""
    parser_descr = 'Create a microstructure and unstructured grid using the'
    parser_descr += ' MicroStructPy package. This is the command line'
    parser_descr += ' interface (CLI) for the package. It reads either demo'
    parser_descr += ' input files or user-created input files and generates:'
    parser_descr += ' 1 A list of seed geometries,'
    parser_descr += ' 2 A polygon/polyhedron mesh, and'
    parser_descr += ' 3. A triangle/tetrahedron mesh.'
    parser_descr += 'Additional outputs, such as plots, can be specified in'
    parser_descr += ' the input file. Please read the input file section of'
    parser_descr += ' the MicroStructPy documentation for more details.'
    parser = argparse.ArgumentParser(description=parser_descr)

    user_descr = 'user-created XML input files, separated by spaces'
    parser.add_argument('user_files', metavar='u', type=str, nargs='*',
                        default=[], help=user_descr)

    demo_descr = 'demonstration names, separated by spaces '
    demo_descr += ' (see documentation for list of available names)'
    parser.add_argument('--demo', dest='demo_files', metavar='d', type=str,
                        nargs='*', default=[], help=demo_descr)

    args = parser.parse_args()

    # run user-generated files
    user_files = [f for fnames in args.user_files for f in glob.glob(fnames)]
    for filename in set(user_files):
        run_file(filename)

    # run demo files
    cli_path = os.path.dirname(__file__)
    demo_path = os.path.normpath(os.path.join(cli_path, 'examples'))
    demo_files = args.demo_files

    if len(demo_files) == 1 and demo_files[0].lower().strip() == 'all':
        xml_files = os.path.normpath(os.path.join(demo_path, '*.xml'))
        xml_demos = [os.path.basename(f) for f in glob.glob(xml_files)]

        py_files = os.path.normpath(os.path.join(demo_path, '*.py'))
        py_demos = [os.path.basename(f) for f in glob.glob(py_files)]

        demo_files = xml_demos + py_demos

    for demo_basename in demo_files:
        ex_file = os.path.join(demo_path, demo_basename)
        ext = demo_basename.split('.')[-1]
        if os.path.exists(ex_file) and (ext in ['py', 'xml']):
            print('Running ' + repr(demo_basename))

            local_filename = os.path.join('./', demo_basename)
            needed_files = _misc.demo_needs.get(demo_basename, [])

            for basename in [demo_basename] + needed_files:
                filename = os.path.join(demo_path, basename)
                try:
                    shutil.move(filename, './')
                except shutil.Error:
                    pass
                else:
                    shutil.copy(basename, filename)

            if ext == 'py':
                subprocess.call(['python', local_filename])
            elif ext == 'xml':
                run_file(local_filename)
            else:
                pass

        else:
            p_str = 'Skipping file ' + repr(demo_basename) + ', since '
            p_str += 'it is not a valid example filename.'
            print(p_str)


def run_file(filename):
    """Run an input file

    This function reads an input file and runs it through the standard
    workflow.

    Args:
        filename (str): The name of an XML input file.

    """
    in_data = read_input(filename)
    phases = in_data['material']
    domain = in_data['domain']
    kwargs = in_data['settings']
    run(phases, domain, **kwargs)


def read_input(filename):
    """Convert input file to dictionary

    This function reads an input file and parses it into a dictionary.

    Args:
        filename (str): The name of an XML input file.

    Returns:
        collections.OrderedDict: Dictionary of run inputs.

    """
    # Read in the file
    file_path = os.path.dirname(filename)

    with open(filename, 'r') as file:
        file_dict = xmltodict.parse(file.read())

    assert 'input' in file_dict, 'Root <input> not found in input file.'
    in_data = dict_convert(file_dict['input'], file_path)

    k_args = ('material', 'domain')
    for arg in k_args:
        assert arg in in_data

    # Get domain
    domain_data = in_data['domain']
    domain_shape = domain_data['shape']
    domain_kwargs = {k: v for k, v in domain_data.items() if k != 'shape'}
    domain = geometry.factory(domain_shape, **domain_kwargs)
    in_data['domain'] = domain

    # Default settings
    kwargs = in_data.get('settings', {})
    run_dir = kwargs.get('directory', '.')
    if not os.path.isabs(run_dir):
        rel_path = os.path.join(file_path, run_dir)
        kwargs['directory'] = os.path.expanduser(os.path.normpath(rel_path))
    in_data['settings'] = kwargs
    return in_data


def run(phases, domain, verbose=False, restart=True, directory='.',
        filetypes={}, rng_seeds={}, plot_axes=True, rtol='fit', edge_opt=False,
        edge_opt_n_iter=100,
        mesh_max_volume=float('inf'), mesh_min_angle=0,
        mesh_max_edge_length=float('inf'), verify=False, color_by='material',
        colormap='viridis', seeds_kwargs={}, poly_kwargs={}, tri_kwargs={}):
    r"""Run MicroStructPy

    This is the primary run function for the package. It performs these steps:

        * Create a list of un-positioned seeds
        * Position seeds in domain
        * Create a polygon mesh from the seeds
        * Create a triangle mesh from the polygon mesh
        * (optional) Perform mesh verification

    Args:
        phases (list or dict): Single phase dictionary or list of multiple
            phase dictionaries. See :ref:`phase_dict_guide` for more details.
        domain (from :mod:`microstructpy.geometry`): The geometry of the
            domain.
        verbose (bool): *(optional)* Option to run in verbose mode.
            Prints status updates to the terminal. Defaults to False.
        restart (bool): *(optional)* Option to run in restart mode.
            Saves caches at the end of each step and reads caches to restart
            the analysis. Defaults to True.
        directory (str): *(optional)* File path where outputs will be saved.
            This path can either be relative to the current directory,
            or an absolute path. Defaults to the current working directory.
        filetypes (dict): *(optional)* Filetypes for the output files.
            A dictionary containing many of the possible file types is::

                filetypes = {'seeds': 'txt',
                             'seeds_plot': ['eps',
                                            'pdf',
                                            'png',
                                            'svg'],
                             'poly': ['txt', 'ply', 'vtk'],
                             'poly_plot': 'png',
                             'tri': ['txt', 'abaqus', 'vtk'],
                             'tri_plot': ['png', 'pdf'],
                             'verify_plot': 'pdf'
                             }

            If an entry is not included in the dictionary, then that output
            is not saved. Default is an empty dictionary. If *restart* is
            True, then 'txt' is added to the 'seeds', 'poly', and 'tri' fields.
        rng_seeds (dict): *(optional)* The random number generator (RNG) seeds.
            The dictionary values should all be non-negative integers.
            Specifically, RNG seeds should be convertible to NumPy `uint32`_.
            An example dictionary is::

                rng_seeds = {'fraction': 0,
                             'phase': 134092,
                             'position': 1,
                             'size': 95,
                             'aspect_ratio': 2,
                             'orienation': 2
                             }

            If a seed is not specified, the default value is 0.
        rtol (float or str): *(optional)* The relative overlap tolerance
            between seeds. This parameter should be between 0 and 1.
            The condition for two circles to overlap is:

            .. math::

                || x_2 - x_1 || + \text{rtol} \min(r_1, r_2) < r_1 + r_2

            The default value is ``'fit'``, which uses the mean and variance
            of the size distribution to estimate a value for rtol.
        edge_opt (bool): *(optional)* This option will maximize the minimum
            edge length in the PolyMesh. The seeds associated with the
            shortest edge are displaced randomly to find improvement and
            this process iterates until `n_iter` attempts have been made
            for a given edge. Defaults to False.
        edge_opt_n_iter (int): *(optional)* Maximum number of iterations per
            edge during optimization. Ignored if `edge_opt` set to False.
            Defaults to 100.
        mesh_max_volume (float): *(optional)* The maximum volume (area in 2D)
            of a mesh cell in the triangular mesh. Default is infinity,
            which turns off the maximum volume quality setting.
            Value should be strictly positive.
        mesh_min_angle (float): *(optional)* The minimum interior angle,
            in degrees,  of a cell in the triangular mesh. For 3D meshes,
            this is the dihedral angle between faces of the tetrahedron.
            Defaults to 0, which turns off the angle quality constraint.
            Value should be in the range 0-60.
        mesh_max_edge_length (float): *(optional)* The maximum edge length of
            elements along grain boundaries. Currently only supported in 2D.
        plot_axes (bool): *(optional)* Option to show the axes in output plots.
            When False, the plots are saved without axes and very tight
            borders. Defaults to True.
        verify (bool): *(optional)* Option to verify the output mesh against
            the input phases. Defaults to False.
        color_by (str): *(optional)* {'material' | 'seed number' |
            'material number'} Option to choose how the polygons/polyhedra
            are colored. Defaults to 'material'.
        colormap (str): *(optional)* Name of the matplotlib colormap to color
            the seeds. Ignored if `color_by='material'`. Defaults to 'viridis',
            the standard matplotlib colormap.
            See `Choosing Colormaps in Matplotlib`_ for more details.
        seed_kwargs (dict): additional keyword arguments that will be passed to
            :meth:`.SeedList.plot`.
        poly_kwargs (dict): Additional keyword arguments that will be passed to
            :meth:`.PolyMesh.plot_facets` in 2D and
            :meth:`.PolyMesh.plot` in 3D.
        tri_kwargs (dict): Additional keyword arguments that will be passed to
            :meth:`.TriMesh.plot`.

    .. _`Specifying Colors`: https://matplotlib.org/users/colors.html
    .. _`Choosing Colormaps in Matplotlib`: https://matplotlib.org/tutorials/colors/colormaps.html
    .. _`uint32`: https://docs.scipy.org/doc/numpy-1.13.0/user/basics.types.html
    """  # NOQA: E501
    # ----------------------------------------------------------------------- #
    # Condition Inputs                                                        #
    # ----------------------------------------------------------------------- #
    # Phases
    # ------
    if type(phases) is not list:
        phases = [phases]

    # Settings
    # --------
    # filetypes
    if restart:
        for kw in ('seeds', 'poly', 'tri'):
            if kw not in filetypes:
                filetypes[kw] = 'txt'
            elif isinstance(filetypes[kw], list):
                if 'txt' not in filetypes[kw]:
                    filetypes[kw].append('txt')
            else:
                filetypes[kw] = [filetypes[kw], 'txt']

    if verbose:
        print('Running MicroStructPy in verbose mode.')

    # ----------------------------------------------------------------------- #
    # Create Directory                                                        #
    # ----------------------------------------------------------------------- #
    if not os.path.exists(directory):
        if verbose:
            print('Creating run directory, ' + directory)
        os.makedirs(directory)

    # ----------------------------------------------------------------------- #
    # Create Seeds                                                            #
    # ----------------------------------------------------------------------- #
    seed_basename = 'seeds.txt'
    seed_filename = os.path.join(directory, seed_basename)
    if restart and os.path.exists(seed_filename):
        # Read seeds from file
        if verbose:
            print('Reading seeds from file.')
            print('Seed file: ' + seed_filename)
        seeds = seeding.SeedList.from_file(seed_filename)
        seeds_created = False
    else:
        seeds_created = True
        # Initialize seeds
        if verbose:
            print('Creating un-positioned list of seeds.')

        seeds = _unpositioned_seeds(phases, domain, rng_seeds)

        if verbose:
            print('There are ' + str(len(seeds)) + ' seeds.')
            print('Positioning seeds in domain.')

        kw = 'position'
        rng_seed = rng_seeds.get(kw, 0)
        pos_dists = {i: p[kw] for i, p in enumerate(phases) if kw in p}
        seeds.position(domain, pos_dists, rng_seed, rtol=rtol, verbose=verbose)

    # Write seeds
    seeds_types = filetypes.get('seeds', [])
    if type(seeds_types) != list:
        seeds_types = [seeds_types]
    for seeds_type in seeds_types:
        fname = seed_filename.rstrip('.txt') + '.' + seeds_type
        if seeds_created or not os.path.exists(fname):
            seeds.write(fname, format=seeds_type)

    # ----------------------------------------------------------------------- #
    # Plot Seeds                                                              #
    # ----------------------------------------------------------------------- #
    plot_types = filetypes.get('seeds_plot', ['png'])
    if not plot_types:
        plot_types = []
    elif type(plot_types) is not list:
        plot_types = [plot_types]

    plot_files = []
    for ext in plot_types:
        fname = os.path.join(directory, 'seeds.' + str(ext))
        if seeds_created or not os.path.exists(fname):
            plot_files.append(fname)

    if plot_files and verbose:
        print('Plotting seeds.')

    plot_seeds(seeds, phases, domain, plot_files, plot_axes, color_by,
               colormap, **seeds_kwargs)

    # ----------------------------------------------------------------------- #
    # Create Polygon Mesh                                                     #
    # ----------------------------------------------------------------------- #
    poly_basename = 'polymesh.txt'
    poly_filename = os.path.join(directory, poly_basename)
    if restart and os.path.exists(poly_filename) and not seeds_created:
        # Read polygon mesh from file
        if verbose:
            print('Reading polygon mesh from file.')
            print('Polygon mesh filename: ' + poly_filename)

        pmesh = PolyMesh.from_file(poly_filename)
        poly_created = False
    else:
        poly_created = True

        # Create polygon mesh from seeds
        if verbose:
            print('Creating polygon mesh.')

        pmesh = PolyMesh.from_seeds(seeds, domain, edge_opt, edge_opt_n_iter,
                                    verbose)

    # Write polymesh
    poly_types = filetypes.get('poly', [])
    if type(poly_types) != list:
        poly_types = [poly_types]

    for poly_type in poly_types:
        fname = poly_filename.replace('.txt', '.' + poly_type)
        if poly_created or not os.path.exists(fname):
            pmesh.write(fname, poly_type)

    # ----------------------------------------------------------------------- #
    # Plot Polygon Mesh                                                       #
    # ----------------------------------------------------------------------- #
    plot_types = filetypes.get('poly_plot', ['png'])
    if not plot_types:
        plot_types = []
    elif type(plot_types) is not list:
        plot_types = [plot_types]

    plot_files = []
    for ext in plot_types:
        fname = os.path.join(directory, 'polymesh.' + str(ext))
        if poly_created or not os.path.exists(fname):
            plot_files.append(fname)

    if plot_files and verbose:
        print('Plotting polygon mesh.')

    plot_poly(pmesh, phases, plot_files, plot_axes, color_by, colormap,
              **poly_kwargs)

    # ----------------------------------------------------------------------- #
    # Create Triangular Mesh                                                  #
    # ----------------------------------------------------------------------- #
    tri_basename = 'trimesh.txt'
    tri_filename = os.path.join(directory, tri_basename)
    exts = {'abaqus': '.inp', 'txt': '.txt', 'str': '.txt', 'tet/tri': '',
            'vtk': '.vtk'}

    if restart and os.path.exists(tri_filename) and not poly_created:
        # Read triangle mesh
        if verbose:
            print('Reading triangular mesh.')
            print('Triangular mesh filename: ' + tri_filename)

        tmesh = TriMesh.from_file(tri_filename)
        tri_created = False
    else:
        tri_created = True
        # Create triangular mesh
        if verbose:
            print('Creating triangular mesh.')

        tmesh = TriMesh.from_polymesh(pmesh, phases, mesh_min_angle,
                                      mesh_max_volume, mesh_max_edge_length)

    # Write triangular mesh
    tri_types = filetypes.get('tri', [])
    if type(tri_types) != list:
        tri_types = [tri_types]

    for tri_type in tri_types:
        fname = tri_filename.replace('.txt', exts[tri_type])
        if tri_created or not os.path.exists(fname):
            tmesh.write(fname, tri_type, seeds, pmesh)

    # ----------------------------------------------------------------------- #
    # Plot Triangular Mesh                                                    #
    # ----------------------------------------------------------------------- #
    plot_types = filetypes.get('tri_plot', ['png'])
    if not plot_types:
        plot_types = []
    elif type(plot_types) is not list:
        plot_types = [plot_types]

    plot_files = []
    for ext in plot_types:
        fname = os.path.join(directory, 'trimesh.' + str(ext))
        if tri_created or not os.path.exists(fname):
            plot_files.append(fname)

    if plot_files and verbose:
        print('Plotting triangular mesh.')

    plot_tri(tmesh, phases, seeds, pmesh, plot_files, plot_axes, color_by,
             colormap, **tri_kwargs)

    # ----------------------------------------------------------------------- #
    # Perform Verification                                                    #
    # ----------------------------------------------------------------------- #
    if not verify:
        if verbose:
            print('Finished running MicroStructPy.')
        return

    if verbose:
        print('Performing verification.')

    # Set, create verification results directory
    ver_dir = os.path.join(directory, 'verification')
    if not os.path.exists(ver_dir):
        os.makedirs(ver_dir)

    # Set plot types
    plottypes = filetypes.get('verify_plot', ['png'])
    if type(plottypes) is not list:
        plottypes = [plottypes]

    # Verify volume fractions
    vol_fracs = verification.volume_fractions(pmesh, len(phases))

    vf_fname = os.path.join(ver_dir, 'volume_fractions.txt')
    verification.write_volume_fractions(vol_fracs, phases, vf_fname)

    fnames = [os.path.join(ver_dir, 'volume_fractions.' + ext) for
              ext in plottypes]
    verification.plot_volume_fractions(vol_fracs, phases, fnames)

    # Screen out certain seeds on the boundaries
    n_dim = domain.n_dim
    default_shape = {2: 'circle', 3: 'sphere'}[n_dim]
    verif_mask = np.full(len(seeds), True)
    for neighbors in pmesh.facet_neighbors:
        if min(neighbors) < 0:
            cell_num = max(neighbors)
            seed_num = pmesh.seed_numbers[cell_num]
            phase_num = pmesh.phase_numbers[cell_num]
            phase = phases[phase_num]
            phase_geom = phase.get('shape', default_shape)
            if phase_geom in ['square', 'rectangle']:
                verif_mask[seed_num] = False

    # Determine seeds of best fit
    fit_seeds = verification.seeds_of_best_fit(seeds, phases, pmesh, tmesh)

    filename = os.path.join(ver_dir, 'fit_seeds.txt')
    fit_seeds.write(filename)

    # Create distribution plots
    verification.plot_distributions(fit_seeds, phases, ver_dir, plottypes,
                                    pmesh, verif_mask)

    # Write Maximum Likelihood Estimates
    mle_phases = verification.mle_phases(fit_seeds, phases, pmesh, verif_mask)

    mle_fname = os.path.join(ver_dir, 'mles.txt')
    verification.write_mle_phases(phases, mle_phases, mle_fname)

    # Write Error Statistics
    error_stats = verification.error_stats(fit_seeds, seeds, phases, pmesh,
                                           verif_mask)

    err_fname = os.path.join(ver_dir, 'err_stats.txt')
    verification.write_error_stats(error_stats, phases, err_fname)

    if verbose:
        print('Finished running MicrostructPy.')


# --------------------------------------------------------------------------- #
#                                                                             #
# Created Unpositioned List of Seeds                                          #
#                                                                             #
# --------------------------------------------------------------------------- #
def _unpositioned_seeds(phases, domain, rng_seeds={}):
    if domain.n_dim == 2:
        dom_vol = domain.area
    else:
        dom_vol = domain.volume
    return seeding.SeedList.from_info(phases, dom_vol, rng_seeds)


# --------------------------------------------------------------------------- #
#                                                                             #
# Plot Seeds                                                                  #
#                                                                             #
# --------------------------------------------------------------------------- #
def plot_seeds(seeds, phases, domain, plot_files=[], plot_axes=True,
               color_by='material', colormap='viridis', **edge_kwargs):
    """Plot seeds

    This function creates formatted plots of a :class:`.SeedList`.

    Args:
        seeds (SeedList): Seed list to plot.
        phases (list): List of phase dictionaries. See :ref:`phase_dict_guide`
            for more details.
        domain (from :mod:`microstructpy.geometry`): Domain geometry.
        plot_files (list): *(optional)* List of files to save the output plot.
            Defaults to saving the plot to ``polymesh.png``.
        plot_axes (bool): *(optional)* Flag to turn the axes on or off.
            True shows the axes, False removes them. Defaults to True.
        color_by (str): *(optional)* {'material' | 'seed number' |
            'material number'} Option to choose how the polygons/polyhedra
            are colored. Defaults to 'material'.
        colormap (str): *(optional)* Name of the matplotlib colormap to color
            the seeds. Ignored if `color_by='material'`. Defaults to 'viridis',
            the standard matplotlib colormap.
            See `Choosing Colormaps in Matplotlib`_ for more details.
        **edge_kwargs: additional keyword arguments that will be passed to
            :meth:`.SeedList.plot`.

    """
    if not plot_files:
        return

    phase_names = []
    given_names = False
    for i, phase in enumerate(phases):
        if 'name' in phase:
            given_names = True
        name = phase.get('name', 'Material ' + str(i + 1))
        phase_names.append(name)

    seed_colors = _seed_colors(seeds, phases, color_by, colormap)
    n_dim = seeds[0].geometry.n_dim

    # Set up axes
    plt.clf()
    plt.close('all')
    fig = plt.figure()
    ax = fig.gca(projection={2: None, 3: Axes3D.name}[n_dim], label='seeds')

    if not plot_axes:
        if n_dim == 2:
            ax.set_axis_off()
            ax.get_xaxis().set_visible(False)
            ax.get_yaxis().set_visible(False)
        else:
            ax._axis3don = False

    # Plot seeds
    edge_kwargs.setdefault('edgecolors', {2: 'k', 3: 'none'}[n_dim])
    if given_names and color_by == 'material':
        seeds.plot(material=phase_names, facecolors=seed_colors, loc=4,
                   **edge_kwargs)
    else:
        seeds.plot(facecolors=seed_colors, **edge_kwargs)

    # Crop to Domain
    d_lims = domain.limits
    if n_dim == 2:
        plt.axis('square')
        plt.xlim(d_lims[0])
        plt.ylim(d_lims[1])
    elif n_dim == 3:
        plt.gca().set_xlim(d_lims[0])
        plt.gca().set_ylim(d_lims[1])
        plt.gca().set_zlim(d_lims[2])

        _misc.axisEqual3D(plt.gca())

    # Save plot
    for fname in plot_files:
        if n_dim == 3:
            fig.subplots_adjust(**_misc.plt_3d_adj)
            plt.savefig(fname, bbox_inches='tight', pad_inches=0.15)
        else:
            plt.savefig(fname, bbox_inches='tight', pad_inches=0)

    plt.close('all')


def _seed_colors(seeds, phases, color_by='material', colormap='viridis'):
    if color_by == 'material':
        return [_phase_color(s.phase, phases) for s in seeds]
    elif color_by == 'seed number':
        n = len(seeds)
        return [_cm_color(i / (n - 1), colormap) for i in range(n)]
    elif color_by == 'material number':
        n = len(phases)
        return [_cm_color(s.phase / (n - 1), colormap) for s in seeds]


def _phase_color(i, phases):
    return phases[i].get('color', 'C' + str(i % 10))


def _phase_color_by(i, phases, color_by='material', colormap='viridis'):
    if color_by == 'material':
        return phases[i].get('color', 'C' + str(i % 10))
    elif color_by == 'material number':
        n = len(phases)
        return _cm_color(i / (n - 1), colormap)


def _cm_color(f, colormap='viridis'):
    return plt.get_cmap(colormap)(f)


# --------------------------------------------------------------------------- #
#                                                                             #
# Plot Polygon                                                                #
#                                                                             #
# --------------------------------------------------------------------------- #
def plot_poly(pmesh, phases, plot_files=['polymesh.png'], plot_axes=True,
              color_by='material', colormap='viridis', **edge_kwargs):
    """Plot polygonal/polyhedral mesh

    This function creates formatted plots of a :class:`.PolyMesh`.

    Args:
        pmesh (PolyMesh): Polygonal mesh to plot.
        phases (list): List of phase dictionaries. See :ref:`phase_dict_guide`
            for more details.
        plot_files (list): *(optional)* List of files to save the output plot.
            Defaults to saving the plot to ``polymesh.png``.
        plot_axes (bool): *(optional)* Flag to turn the axes on or off.
            True shows the axes, False removes them. Defaults to True.
        color_by (str): *(optional)* {'material' | 'seed number' |
            'material number'} Option to choose how the polygons/polyhedra
            are colored. Defaults to 'material'.
        colormap (str): *(optional)* Name of the matplotlib colormap to color
            the seeds. Ignored if `color_by='material'`. Defaults to 'viridis',
            the standard matplotlib colormap.
            See `Choosing Colormaps in Matplotlib`_ for more details.
        **edge_kwargs: Additional keyword arguments that will be passed to
            :meth:`.PolyMesh.plot_facets` in 2D and
            :meth:`.PolyMesh.plot` in 3D.

    """
    if not plot_files:
        return

    n_dim = len(pmesh.points[0])

    phase_colors = []
    phase_names = []
    given_names = False
    for i, phase in enumerate(phases):
        color = phase.get('color', 'C' + str(i % 10))
        if 'name' in phase:
            given_names = True
        name = phase.get('name', 'Material ' + str(i + 1))
        phase_colors.append(color)
        phase_names.append(name)

    # Set up axes
    plt.clf()
    plt.close('all')
    fig = plt.figure()
    ax = fig.gca(projection={2: None, 3: Axes3D.name}[n_dim], label='poly')

    if not plot_axes:
        if n_dim == 2:
            ax.set_axis_off()
            ax.get_xaxis().set_visible(False)
            ax.get_yaxis().set_visible(False)
        else:
            ax._axis3don = False

    # Plot polygons
    fcs = _poly_colors(pmesh, phases, color_by, colormap, n_dim)
    if n_dim == 2:
        if given_names and color_by == 'material':
            pmesh.plot(facecolors=fcs, material=phase_names)
        else:
            pmesh.plot(facecolors=fcs)

        edge_color = edge_kwargs.pop('edgecolors', (0, 0, 0, 1))
        facet_colors = []
        for neigh_pair in pmesh.facet_neighbors:
            if facet_check(neigh_pair, pmesh, phases):
                facet_colors.append(edge_color)  # black
            else:
                facet_colors.append('none')

        edge_kwargs.setdefault('capstyle', 'round')
        pmesh.plot_facets(color=facet_colors, index_by='facet', **edge_kwargs)
    else:
        edge_kwargs.setdefault('edgecolors', 'k')
        if given_names and color_by == 'material':
            pmesh.plot(facecolors=fcs, index_by='seed', material=phase_names,
                       **edge_kwargs)
        else:
            pmesh.plot(facecolors=fcs, index_by='seed', **edge_kwargs)

    # save plot
    for fname in plot_files:
        if n_dim == 3:
            fig.subplots_adjust(**_misc.plt_3d_adj)
            plt.savefig(fname, bbox_inches='tight', pad_inches=0.15)
        else:
            plt.savefig(fname, bbox_inches='tight', pad_inches=0)
    plt.close('all')


def _poly_colors(pmesh, phases, color_by, colormap, n_dim):
    if n_dim == 2:
        if color_by == 'material':
            r_colors = [_phase_color(n, phases) for n in pmesh.phase_numbers]
        elif color_by == 'seed number':
            n = max(pmesh.seed_numbers) + 1
            r_colors = [_cm_color(s / (n - 1), colormap) for s in
                        pmesh.seed_numbers]
        elif color_by == 'material number':
            n = len(phases)
            r_colors = [_cm_color(p / (n - 1), colormap) for p in
                        pmesh.phase_numbers]
        n_seeds = max(pmesh.seed_numbers) + 1
        s_colors = ['none' for i in range(n_seeds)]
        for seed_num, r_c in zip(pmesh.seed_numbers, r_colors):
            s_colors[seed_num] = r_c
        return s_colors
    else:
        s2p = {s: p for s, p in zip(pmesh.seed_numbers, pmesh.phase_numbers)}
        n = max(s2p.keys()) + 1
        colors = []
        for s in range(n):
            if color_by == 'material':
                phase_num = s2p[s]
                color = _phase_color(phase_num, phases)
            elif color_by == 'seed number':
                color = _cm_color(s / (n - 1), colormap)
            elif color_by == 'material number':
                n_phases = len(phases)
                color = _cm_color(s2p[s] / (n_phases - 1), colormap)
            else:
                color = 'none'
            colors.append(color)
        return colors


# --------------------------------------------------------------------------- #
#                                                                             #
# Plot Triangular Mesh                                                        #
#                                                                             #
# --------------------------------------------------------------------------- #
def plot_tri(tmesh, phases, seeds, pmesh, plot_files=[], plot_axes=True,
             color_by='material', colormap='viridis', **edge_kwargs):
    """Plot seeds

    This function creates formatted plots of a :class:`.TriMesh`.

    Args:
        tmesh (TriMesh): Triangular mesh to plot.
        phases (list): List of phase dictionaries. See :ref:`phase_dict_guide`
            for more details.
        seeds (SeedList): List of seed geometries.
        pmesh (PolyMesh): Polygonal mesh from which ``tmesh`` was generated.
        plot_files (list): *(optional)* List of files to save the output plot.
            Defaults to saving the plot to ``polymesh.png``.
        plot_axes (bool): *(optional)* Flag to turn the axes on or off.
            True shows the axes, False removes them. Defaults to True.
        color_by (str): *(optional)* {'material' | 'seed number' |
            'material number'} Option to choose how the polygons/polyhedra
            are colored. Defaults to 'material'.
        colormap (str): *(optional)* Name of the matplotlib colormap to color
            the seeds. Ignored if `color_by='material'`. Defaults to 'viridis',
            the standard matplotlib colormap.
            See `Choosing Colormaps in Matplotlib`_ for more details.
        **edge_kwargs: Additional keyword arguments that will be passed to
            :meth:`.TriMesh.plot`.

    """
    if not plot_files:
        return

    n_dim = len(tmesh.points[0])

    phase_colors = []
    phase_names = []
    given_names = False
    for i, phase in enumerate(phases):
        color = phase.get('color', 'C' + str(i % 10))
        name = phase.get('name', 'Material ' + str(i + 1))
        phase_colors.append(color)
        phase_names.append(name)
        if 'name' in phase:
            given_names = True

    # Set up axes
    plt.clf()
    plt.close('all')
    fig = plt.figure()
    ax = fig.gca(projection={2: None, 3: Axes3D.name}[n_dim], label='tri')

    if not plot_axes:
        if n_dim == 2:
            ax.set_axis_off()
            ax.get_xaxis().set_visible(False)
            ax.get_yaxis().set_visible(False)
        else:
            ax._axis3don = False

    # Determine which facets are visible
    vis_regions = set()
    invis_regions = set(range(-6, 0))
    f_front = set([i for i, fn in enumerate(pmesh.facet_neighbors)
                   if min(fn) < 0])
    while f_front:
        new_front = set()
        for f in f_front:
            neighs = set(pmesh.facet_neighbors[f])
            for n in neighs - invis_regions:
                p = pmesh.phase_numbers[n]
                p_type = phases[p].get('material_type', 'solid')
                if p_type in _misc.kw_void:
                    new_front |= set(pmesh.regions[n])
                else:
                    vis_regions.add(n)
        new_front -= f_front
        f_front = new_front

    # Determine facet colors based on visibility
    seed_colors = _seed_colors(seeds, phases, color_by, colormap)
    facet_colors = []
    facet_phases = []
    for i, fn in enumerate(pmesh.facet_neighbors):
        if _f_plottable(fn, vis_regions, invis_regions):
            r = list(set(fn) - invis_regions)[0]
            s = pmesh.seed_numbers[r]
            color = seed_colors[s]
            phase = seeds[s].phase
        else:
            color = 'none'
            phase = -1
        facet_colors.append(color)
        facet_phases.append(phase)

    # plot triangle mesh
    edge_kwargs.setdefault('linewidths', {2: 0.5, 3: 0.1}[n_dim])
    edge_kwargs.setdefault('edgecolors', 'k')
    if color_by in ('material', 'material number'):
        n = len(phases)
        if n_dim == 2:
            cs = seed_colors
        else:
            cs = facet_colors
            cs.append('none')

        cs = [_phase_color_by(i, phases, color_by, colormap) for i in range(n)]
        cs.append('none')

        old_e_att = np.copy(tmesh.element_attributes)
        old_f_att = np.copy(tmesh.facet_attributes)
        tmesh.element_attributes = [seeds[i].phase for i in old_e_att]
        tmesh.facet_attributes = [facet_phases[a] for a in old_f_att]

        if given_names:
            tmesh.plot(facecolors=cs, index_by='attribute',
                       material=phase_names, **edge_kwargs)
        else:
            tmesh.plot(facecolors=cs, index_by='attribute', **edge_kwargs)

        tmesh.element_attributes = old_e_att
        tmesh.facet_attributes = old_f_att
    else:
        tmesh.plot(facecolors=facet_colors, index_by='attribute',
                   **edge_kwargs)

    # save plot
    for fname in plot_files:
        if n_dim == 3:
            fig.subplots_adjust(**_misc.plt_3d_adj)
            plt.savefig(fname, bbox_inches='tight', pad_inches=0.15)
        else:
            plt.savefig(fname, bbox_inches='tight', pad_inches=0)

    plt.close('all')


def _f_plottable(n_pair, vis, invis):
    if set(n_pair) <= vis or set(n_pair) <= invis:
        return False
    return True


# --------------------------------------------------------------------------- #
#                                                                             #
# Recursively Convert Dictionary Contents                                     #
#                                                                             #
# --------------------------------------------------------------------------- #
def dict_convert(dictionary, filepath='.'):
    """Convert dictionary from xmltodict_

    The xmltodict_ ``parse`` method creates dictionaries with values that
    are all strings, rather than strings, floats, ints, etc.
    This function recursively searches the dictionary for string values and
    attempts to convert the dictionary values.

    If a dictionary contains the key ``dist_type``, it is assumed that
    the corresponding name is a :mod:`scipy.stats` statistical distribution
    and the remaining keys are inputs for that distribution,
    with two exceptions.
    First, if the value of ``dist_type`` is ``cdf``, then the remaining key
    should be ``filename`` and its value should be the path to a CSV file,
    where each row contains the (x, CDF) points along the CDF curve.
    Second, if the value of ``dist_type`` is ``histogram``, then the remaining
    key should also be ``filename`` and its value should be the path to a CSV
    file.
    For the histogram, the first row of this CDF should be the *n* bin heights
    and the second row should be the *n+1* bin locations.

    Additionally, if a key in the dictionary contains ``filename`` or
    ``directory`` and the value associated with that key is a relative path,
    then the filepath is converted from a relative to an absolute path using
    the ``filepath`` input as the reference point.
    This behavior can be switched off by setting ``filepath=False``.

    Args:
        dictionary (list, dict, or collections.OrderedDict): Dictionary or
            dictionaries to be converted.
        filepath (str): *(optional)* Reference path to resolve relative paths.

    Returns:
        list or collections.OrderedDict: A copy of the input where the string
        values have been converted. If only one dict is passed into the
        function, then an instance of :class:`collections.OrderedDict` is
        returned.

    .. _xmltodict: https://github.com/martinblech/xmltodict
    """

    # Convert lists
    if isinstance(dictionary, list):
        return [dict_convert(d) for d in dictionary]

    # Convert strings
    if isinstance(dictionary, str):
        return _misc.from_str(dictionary)

    # Convert Nones
    if dictionary is None:
        return {}

    # Convert filepaths
    for key in dictionary:
        val = dictionary[key]
        if any([s in key.lower() for s in ('filename', 'directory')]):
            if not os.path.isabs(val) and filepath:
                new_val = os.path.abspath(os.path.join(filepath, val))
            else:
                new_val = val
            dictionary[key] = new_val

    # Convert SciPy.stats distributions
    if 'dist_type' in dictionary:
        return _dist_convert(dictionary)

    # Convert Dictionaries
    new_dict = collections.OrderedDict()
    for key in dictionary:
        val = dictionary[key]
        new_val = dict_convert(val, filepath)
        new_dict[key] = new_val
    return new_dict


def _dist_convert(dist_dict):
    """Convert distribution dictionary to distribution"""

    dist_type = dist_dict['dist_type'].strip().lower()
    params = {k: _misc.from_str(v) for k, v in dist_dict.items()}
    del params['dist_type']

    if dist_type == 'cdf':
        cdf_filename = params['filename']
        with open(cdf_filename, 'r') as file:
            cdf = [[float(s) for s in line.split(',')] for line in file]

        bin_bnds = [x for x, _ in cdf]
        bin_cnts = [cdf[i + 1][1] - cdf[i][1] for i in range(len(cdf) - 1)]
        return scipy.stats.rv_histogram(tuple([bin_cnts, bin_bnds]))

    elif dist_type == 'histogram':
        hist_filename = params['filename']
        with open(hist_filename, 'r') as file:
            hist = [[float(s) for s in line.split(',')] for line in file]
        return scipy.stats.rv_histogram(tuple(hist))

    else:
        return scipy.stats.__dict__[dist_type](**params)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function

from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

from setuptools import find_packages
from setuptools import setup

desc = 'Microstructure modeling, mesh generation, analysis, and visualization.'


def read(*fname):
    return open(join(dirname(__file__), *fname)).read()


def find_version(*fname):
    ver_str = ''
    for line in read(*fname).split('\n'):
        if line.startswith('__version__') and '=' in line:
            ver_str = line.split('=')[-1].strip().strip('\"').strip('\'')
            break
    return ver_str


setup(
    name='microstructpy',
    version=find_version('src', 'microstructpy', '__init__.py'),
    license='MIT License',
    description=desc,
    long_description=read('README.rst'),
    long_description_content_type=' text/x-rst',
    author='Kenneth (Kip) Hart',
    author_email='kiphart91@gmail.com',
    url='https://github.com/kip-hart/MicroStructPy',
    project_urls={
        'Documentation': 'https://docs.microstructpy.org',
    },
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    package_data={'': ['src/microstructpy/examples/*.{csv,py,xml}',
                       'src/microstructpy/examples/aluminum_micro.png']},
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list:
        # http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Physics'
    ],
    keywords=[
        'microstructure',
        'micromechanics',
        'finite element',
        'FEM', 'FEA',
        'mesh',
        'polycrystal',
        'tessellation',
        'Laguerre tessellation',
        'multi-sphere'
    ],
    install_requires=[
        'aabbtree>=2.5.0',
        'matplotlib>=3.0.0',
        'pybind11',  # must come before meshpy for successful install
        'lsq-ellipse',
        'meshpy>=2018.2.1',
        'numpy>=1.13.0',
        'pyquaternion',
        'pyvoro-mmalahe',  # install issue with pyvoro
        'scipy',
        'xmltodict'
    ],
    entry_points={
        'console_scripts': [
            'microstructpy = microstructpy.cli:main',
        ]
    },
)



# --------------------------------------------------------------------------- #
#                                                                             #
# Import Modules                                                              #
#                                                                             #
# --------------------------------------------------------------------------- #
from __future__ import division

import warnings

import numpy as np
import scipy.spatial.distance
from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from pyquaternion import Quaternion

from microstructpy import _misc
from microstructpy.geometry.ellipse import Ellipse

__author__ = 'Kenneth (Kip) Hart'


# --------------------------------------------------------------------------- #
#                                                                             #
# Ellipsoid Class                                                             #
#                                                                             #
# --------------------------------------------------------------------------- #
class Ellipsoid(object):
    """A 3D Ellipsoid

    This class contains the data and functions for a 3D ellispoid.
    It is defined by its center, axes, and orientation.

    If multiple keywords are given for the shape of the ellipsoid, there
    is no guarantee for which keywords are used.

    Args:
        a (float): *(optional)* First semi-axis of ellipsoid. Default is 1.
        b (float): *(optional)* Second semi-axis of ellipsoid. Default is 1.
        c (float): *(optional)* Third semi-axis of ellipsoid. Default is 1.
        center (list): *(optional)* The ellipsoid center.
            Defaults to (0, 0, 0).
        axes (list): *(optional)* List of 3 semi-axes.
            Defaults to (1, 1, 1).
        size (float): *(optional)* The diameter of a sphere with equal volume.
            Defaults to 2.
        ratio_ab (float): *(optional)* The ratio of a to b.
        ratio_ac (float): *(optional)* The ratio of a to c.
        ratio_bc (float): *(optional)* The ratio of b to c.
        ratio_ba (float): *(optional)* The ratio of b to a.
        ratio_ca (float): *(optional)* The ratio of c to a.
        ratio_cb (float): *(optional)* The ratio of c to b.
        rot_seq (list): *(optional)* List of rotations (deg). Each element of
            the list should be an (axis, angle) tuple. The options for the
            axis are: 'x', 'y', 'z', 1, 2, or 3.
            For example::

                rot_seq = [('x', 10), (2, -20), ('z', 85), ('x', 21)]

        rot_seq_deg (list): *(optional)* Alias for ``rot_seq``, with degrees
            stated explicitly.
        rot_seq_rad (list): *(optional)* Same format as ``rot_seq``, except the
            angles are expressed in radians.
        matrix (numpy.ndarray): *(optional)* A 3x3 rotation matrix expressing
            the orientation of the ellipsoid. Defaults to the identity.
        position : *(optional)* Alias for ``center``.
        orientation: *(optional)* Alias for ``matrix``.
    """
    def __init__(self, **kwargs):
        # Position
        if 'center' in kwargs:
            self.center = kwargs['center']

        elif 'position' in kwargs:
            self.center = kwargs['position']

        else:
            self.center = (0, 0, 0)

        # Axes
        self.a = None
        self.b = None
        self.c = None

        ratio_ab = None
        ratio_ac = None
        ratio_bc = None

        size = None

        if 'a' in kwargs:
            self.a = kwargs['a']
        if 'b' in kwargs:
            self.b = kwargs['b']
        if 'c' in kwargs:
            self.c = kwargs['c']
        if 'ratio_ab' in kwargs:
            ratio_ab = kwargs['ratio_ab']
        elif 'ratio_ba' in kwargs:
            ratio_ab = 1 / kwargs['ratio_ba']
        if 'ratio_ac' in kwargs:
            ratio_ac = kwargs['ratio_ac']
        elif 'ratio_ca' in kwargs:
            ratio_ac = 1 / kwargs['ratio_ca']
        if 'ratio_bc' in kwargs:
            ratio_bc = kwargs['ratio_bc']
        elif 'ratio_cb' in kwargs:
            ratio_bc = 1 / kwargs['ratio_cb']
        if 'size' in kwargs:
            size = kwargs['size']
        elif 'volume' in kwargs:
            size = 2 * np.cbrt(3 * kwargs['volume'] / (4 * np.pi))
        if 'axes' in kwargs:
            self.a, self.b, self.c = kwargs['axes']

        if (self.a is not None) and (self.b is not None):
            ratio_ab = self.a / self.b
        if (self.a is not None) and (self.c is not None):
            ratio_ac = self.a / self.c
        if (self.b is not None) and (self.c is not None):
            ratio_bc = self.b / self.c

        if (ratio_ab is not None) and (ratio_bc is not None):
            ratio_ac = ratio_ab * ratio_bc
        if (ratio_ab is not None) and (ratio_ac is not None):
            ratio_bc = ratio_ac / ratio_ab
        if (ratio_ac is not None) and (ratio_bc is not None):
            ratio_ab = ratio_ac / ratio_bc

        for _ in range(2):
            if self.a is None:
                if (ratio_ab is not None) and (self.b is not None):
                    self.a = ratio_ab * self.b
                elif (ratio_ac is not None) and (self.c is not None):
                    self.a = ratio_ac * self.c

            if self.b is None:
                if (ratio_ab is not None) and (self.a is not None):
                    self.b = self.a / ratio_ab
                elif (ratio_bc is not None) and (self.c is not None):
                    self.b = ratio_bc * self.a

            if self.c is None:
                if (ratio_ac is not None) and (self.a is not None):
                    self.c = self.a / ratio_ac
                elif (ratio_bc is not None) and (self.b is not None):
                    self.c = self.b / ratio_bc

        if (self.a is not None) and (self.b is not None):
            ratio_ab = self.a / self.b
        if (self.a is not None) and (self.c is not None):
            ratio_ac = self.a / self.c
        if (self.b is not None) and (self.c is not None):
            ratio_bc = self.b / self.c

        if (ratio_ab is not None) and (ratio_bc is not None):
            ratio_ac = ratio_ab * ratio_bc
        if (ratio_ab is not None) and (ratio_ac is not None):
            ratio_bc = ratio_ac / ratio_ab
        if (ratio_ac is not None) and (ratio_bc is not None):
            ratio_ab = ratio_ac / ratio_bc

        if (size is not None):
            r_eff = 0.5 * size
            r3 = r_eff * r_eff * r_eff
            if (self.a is not None) and (self.b is not None):
                self.c = r3 / (self.a * self.b)
            elif (self.b is not None) and (self.c is not None):
                self.a = r3 / (self.b * self.c)
            elif (self.a is not None) and (self.c is not None):
                self.b = r3 / (self.a * self.c)
            elif (self.a is not None) and (ratio_bc is not None):
                r2 = r3 / self.a
                # r2 = b * c
                # r2 = c * ratio_bc * c = ratio_bc * c^2
                # c = sqrt(r2 / ratio_bc) and b = ratio_bc * c
                self.c = np.sqrt(r2 / ratio_bc)
                self.b = ratio_bc * self.c
            elif (self.b is not None) and (ratio_ac is not None):
                r2 = r3 / self.b
                self.c = np.sqrt(r2 / ratio_ac)
                self.a = ratio_ac * self.c
            elif (self.c is not None) and (ratio_ab is not None):
                r2 = r3 / self.c
                self.b = np.sqrt(r2 / ratio_ab)
                self.a = ratio_ab * self.b
            else:
                # r3 = a * b * c
                # r3 = a * (a / ratio_ab) * (a / ratio_ac)
                # r3 * ratio_ab * ratio_ac = a^3
                # a = r_eff * cbrt(ratio_ab * ratio_ac)

                self.a = r_eff * np.cbrt(ratio_ab * ratio_ac)
                self.b = self.a / ratio_ab
                self.c = self.a / ratio_ac

        if self.a is None:
            self.a = 1
        if self.b is None:
            self.b = 1
        if self.c is None:
            self.c = 1

        # Orientation
        if 'rot_seq' in kwargs:
            self.rot_seq = kwargs['rot_seq']
        elif 'rot_seq_deg' in kwargs:
            self.rot_seq = kwargs['rot_seq_deg']
        elif 'rot_seq_rad' in kwargs:
            rs_rad = kwargs['rot_seq_rad']
            self.rot_seq = [(ax, 180 * ang / np.pi) for ax, ang in rs_rad]
        elif ('matrix' in kwargs) or ('orientation' in kwargs):
            # adapted from:
            # https://www.learnopencv.com/rotation-matrix-to-euler-angles/
            if 'matrix' in kwargs:
                R = kwargs['matrix']
            else:
                R = kwargs['orientation']
            sy = np.sqrt(R[0][0] * R[0][0] + R[1][0] * R[1][0])

            if sy > 1e-6:
                x = np.arctan2(R[2][1], R[2][2])
                y = np.arctan2(-R[2][0], sy)
                z = np.arctan2(R[1][0], R[0][0])
            else:
                x = np.arctan2(-R[1][2], R[1][1])
                y = np.arctan2(-R[2][0], sy)
                z = 0

            axes = ('z', 'y', 'x')
            angs = [180 * ang / np.pi for ang in (z, y, x)]
            self.rot_seq = list(zip(axes, angs))
        else:
            self.rot_seq = []

    # ----------------------------------------------------------------------- #
    # Ellipsoid of Best Fit                                                   #
    # ----------------------------------------------------------------------- #
    def best_fit(self, points):
        """Find ellipsoid of best fit.

        This function takes a list of 3D points and computes the ellipsoid of
        best fit for the points. It uses a published algorithm to fit the
        ellipsoid, then attempts to define the axes in such a way that they
        most align with this ellipsoid's axes. [#turner]_

        Args:
            points (list): Points to fit ellipsoid

        Returns:
            Ellipsoid: The ellipsoid that best fits the points.

        .. [#turner] Turner, D. A., Anderson, I. J., Mason, J. C., and Cox,
            M. G., "An Algorithm for Fitting an Ellipsoid to Data," *National
            Physical Laboratory*, 1999, The United Kingdom.
            (http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.36.2773&rep=rep1&type=pdf)
        """  # NOQA: E501
        pts = np.array(points)
        pts_mean = pts.mean(axis=0)
        trans_pts = pts - pts_mean

        x, y, z = trans_pts.T

        L = np.zeros((len(x), 9), dtype='float')
        L[:, 0] = x * x + y * y - 2 * z * z
        L[:, 1] = x * x - 2 * y * y + z * z
        L[:, 2] = 4 * x * y
        L[:, 3] = 2 * x * z
        L[:, 4] = 2 * y * z
        L[:, 5] = x
        L[:, 6] = y
        L[:, 7] = z
        L[:, 8] = 1

        e = x * x + y * y + z * z

        u, v, m, n, p, q, r, s, t = np.linalg.lstsq(L, e, rcond=None)[0]
        axx = 1 - u - v
        ayy = 1 - u + 2 * v
        azz = 1 + 2 * u - v
        axy = -4 * m
        axz = -2 * n
        ayz = -2 * p
        ax = - q
        ay = - r
        az = - s
        ac = - t

        hom_mat = np.array([[axx, 0.5 * axy, 0.5 * axz, 0.5 * ax],
                            [0.5 * axy, ayy, 0.5 * ayz, 0.5 * ay],
                            [0.5 * axz, 0.5 * ayz, azz, 0.5 * az],
                            [0.5 * ax, 0.5 * ay, 0.5 * az, ac]])

        cen = np.linalg.lstsq(-hom_mat[:3, :3], hom_mat[-1, :3], rcond=None)[0]
        glbl_cen = cen + pts_mean

        T = np.eye(4)
        T[-1, :3] = cen
        R = T.dot(hom_mat.dot(T.T))
        evals, evecs = np.linalg.eigh(- R[:3, :3] / R[-1, -1])

        if np.any(evals <= 0):
            w_str = 'Some of the eigenvalues of the quadratic form for'
            w_str += 'the ellipsoid are non-positive. Using absolute value, '
            w_str += 'but fit may be poor.'
            warnings.warn(w_str, RuntimeWarning)

        # extract ellipsoid parameters
        try:
            # reorganize the axes
            opt_axes = 1 / np.sqrt(np.abs(evals))
            axes = np.zeros(3)
            avail_axes = np.copy(opt_axes)
            rearr = np.zeros((3, 3))
            for ax_i, ax in enumerate(self.axes):
                opt_i = np.argmin(np.abs(avail_axes - ax))
                rearr[ax_i, opt_i] = 1
                avail_axes[opt_i] = np.inf
                axes[ax_i] = opt_axes[opt_i]
            new_evecs = evecs.dot(rearr)

            # settle the orientation ambiguity
            ori_dot = self.matrix.T.dot(new_evecs)
            vec_flip = np.sign(np.diagonal(ori_dot))
            ori_matrix = new_evecs.dot(np.diag(vec_flip))

        except AttributeError:
            axes = 1 / np.sqrt(np.abs(evals))
            ori_matrix = evecs

        return type(self)(center=glbl_cen, axes=axes, matrix=ori_matrix)

    # ----------------------------------------------------------------------- #
    # String and Representation Functions                                     #
    # ----------------------------------------------------------------------- #
    def __str__(self):
        str_str = 'center: ' + str(tuple(self.center)) + '\n'
        str_str += 'a: ' + str(self.a) + '\n'
        str_str += 'b: ' + str(self.b) + '\n'
        str_str += 'c: ' + str(self.c)
        if len(self.rot_seq) > 0:
            str_str += '\nrot_seq: ('
            for i, (ax, ang) in enumerate(self.rot_seq):
                str_str += '(' + str(ax) + ', ' + str(ang) + ')'
                if i < len(self.rot_seq) - 1:
                    str_str += ', '
                else:
                    str_str += ')'
        return str_str

    def __repr__(self):
        repr_str = 'Ellipsoid('
        repr_str += 'center=' + repr(tuple(self.center))
        repr_str += ', axes=' + repr(tuple([self.a, self.b, self.c]))
        rot_str = repr(tuple([tuple(r) for r in self.rot_seq]))
        repr_str += ', rot_seq=' + rot_str
        repr_str += ')'
        return repr_str

    # ----------------------------------------------------------------------- #
    # Size and Orientation Getters                                            #
    # ----------------------------------------------------------------------- #
    @property
    def size(self):
        """float: diameter of equivalent volume sphere"""
        return 2 * np.cbrt(self.a * self.b * self.c)

    @property
    def axes(self):
        """tuple: the 3 semi-axes of the ellipsoid"""
        return self.a, self.b, self.c

    @property
    def ratio_ab(self):
        """float: ratio of x-axis length to y-axis length"""
        return self.a / self.b

    @property
    def ratio_ba(self):
        """float: ratio of y-axis length to x-axis length"""
        return self.b / self.a

    @property
    def ratio_ac(self):
        """float: ratio of x-axis length to z-axis length"""
        return self.a / self.c

    @property
    def ratio_ca(self):
        """float: ratio of z-axis length to x-axis length"""
        return self.c / self.a

    @property
    def ratio_bc(self):
        """float: ratio of y-axis length to z-axis length"""
        return self.b / self.c

    @property
    def ratio_cb(self):
        """float: ratio of z-axis length to y-axis length"""
        return self.c / self.b

    @property
    def rot_seq_deg(self):
        """list: rotation sequence, with angles in degrees"""
        return self.rot_seq

    @property
    def rot_seq_rad(self):
        """list: rotation sequence, with angles in radiands"""
        return [(ax, np.pi * ang / 180) for ax, ang in self.rot_seq]

    @property
    def matrix(self):
        """numpy.ndarray: A 3x3 rotation matrix"""
        ax_dict = {1: 0, 2: 1, 3: 2, 'x': 0, 'y': 1, 'z': 2}

        q = Quaternion()
        for ax, ang in self.rot_seq_deg:
            vec = np.eye(3)[ax_dict[ax]]
            q *= Quaternion(axis=vec, degrees=ang)
        return q.rotation_matrix

    @property
    def orientation(self):
        """numpy.ndarray: A 3x3 rotation matrix"""
        return self.matrix

    # ----------------------------------------------------------------------- #
    # Quadratic Form Matrix                                                   #
    # ----------------------------------------------------------------------- #
    @property
    def matrix_quadform(self):
        """numpy.ndarray: Matrix of the quadratic form"""
        R = np.array(self.orientation)
        scl_mat = np.diag(1 / (np.array(self.axes) * np.array(self.axes)))
        return R.dot(scl_mat).dot(R.T)

    @property
    def matrix_quadeq(self):
        """numpy.ndarray: Matrix of the quadratic equation"""
        A33 = self.matrix_quadform
        grad_vec = - A33.dot(self.center)
        cen_vec = np.array(self.center).reshape(-1, 1)
        const = cen_vec.T.dot(A33.dot(cen_vec))[0, 0] - 1

        quad_mat = np.zeros((4, 4))
        quad_mat[:3, :3] = A33
        quad_mat[-1, :3] = grad_vec.T
        quad_mat[:3, -1] = grad_vec
        quad_mat[-1, -1] = const
        return quad_mat

    @property
    def coefficients(self):
        """tuple: coeffificients of equation,
        :math:`(A, B, C, D, E, F, G, H, K, L)` in
        :math:`Ax^2 + Bxy + Cxz + Dy^2 + Eyz + Fz^2 + Gx + Hy + Kz + L = 0`
        """
        quad_eq = self.matrix_quadeq
        A = quad_eq[0, 0]
        B = 2 * quad_eq[0, 1]
        C = 2 * quad_eq[0, 2]
        D = quad_eq[1, 1]
        E = 2 * quad_eq[1, 2]
        F = quad_eq[2, 2]
        G = 2 * quad_eq[3, 0]
        H = 2 * quad_eq[3, 1]
        K = 2 * quad_eq[3, 2]
        L = quad_eq[3, 3]
        return A, B, C, D, E, F, G, H, K, L

    # ----------------------------------------------------------------------- #
    # Number of Dimensions                                                    #
    # ----------------------------------------------------------------------- #
    @property
    def n_dim(self):
        """int: number of dimensions, 3"""
        return 3

    # ----------------------------------------------------------------------- #
    # Volume Property                                                         #
    # ----------------------------------------------------------------------- #
    @property
    def volume(self):
        """ float: volume of ellipsoid, :math:`V = \\frac{4}{3}\\pi a b c`"""
        return 4 * np.pi * self.a * self.b * self.c / 3

    @classmethod
    def volume_expectation(cls, **kwargs):
        r"""Expected value of volume.

        This function computes the expected value for the volume of an
        ellipsoid. The keyword arguments are the same as the input parameters
        for the class, :class:`microstructpy.geometry.Ellipsoid`. The
        values for these keywords can be either constants or distributions from
        the SciPy :mod:`scipy.stats` module.

        The expected value is computed by the following formula:

        .. math::

            \mathbb{E}[V] &= \mathbb{E}[\frac{4}{3}\pi A B C] \\
                            &= \frac{4}{3}\pi \mathbb{E}[A] \mathbb{E}[B] \mathbb{E}[C] \\
                            &= \frac{4}{3}\pi \mu_A \mu_B \mu_C

        If the ellisoid is specified by size and aspect ratios, then the
        expected volume is computed by:

        .. math::

            \mathbb{E}[V] &= \mathbb{E}[\frac{\pi}{6} S^3] \\
                            &= \frac{\pi}{6} (\mu_S^3 + 3 \mu_S \sigma_S^2 + \gamma_{1, S} \sigma_S^3)

        If the ellipsoid is specified using a combination of semi-axes and
        aspect ratios, then the expected volume is the mean of 1000 random
        samples:

        .. math::

            \mathbb{E}[V] \approx \frac{1}{n} \sum_{i=1}^n V_i

        where :math:`n=1000`.

        Args:
            **kwargs: Keyword arguments, see :class:`microstructpy.geometry.Ellipsoid`.

        Returns:
            float: Expected value of the volume of the sphere.

        """  # NOQA: E501
        # Check for size distribution
        if 'size' in kwargs:
            s_dist = kwargs['size']
            if type(s_dist) in (float, int):
                return 0.5 * np.pi * s_dist * s_dist * s_dist / 3
            return 0.5 * np.pi * s_dist.moment(3) / 3

        if 'volume' in kwargs:
            v_dist = kwargs['volume']
            try:
                v_exp = v_dist.moment(1)
            except AttributeError:
                v_exp = v_dist
            return v_exp

        # check for a, b, and c distribution
        try:
            exp_vol = 4 * np.pi / 3
            for kw in ('a', 'b', 'c'):
                dist = kwargs[kw]
                try:
                    mu = dist.moment(1)
                except AttributeError:
                    mu = dist
                exp_vol *= mu
            return exp_vol
        except KeyError:
            pass

        # Use Monte Carlo to determine expected volume
        n_trials = 1000
        kws = set(kwargs.keys()) - set(_misc.ori_kws)

        total_vol = 0
        for i in range(n_trials):
            params = {}
            for kw in kws:
                try:
                    params[kw] = kwargs[kw].rvs()
                except AttributeError:
                    params[kw] = kwargs[kw]
            total_vol += Ellipsoid(**params).volume
        avg_vol = total_vol / n_trials
        return avg_vol

    # ----------------------------------------------------------------------- #
    # Bounding Circles                                                        #
    # ----------------------------------------------------------------------- #
    @property
    def bound_max(self):
        """tuple: maximum bounding sphere, (x, y, z, r)"""
        r = max(self.a, self.b, self.c)
        return tuple(list(self.center) + [r])

    @property
    def bound_min(self):
        """tuple: minimum interior sphere, (x, y, z, r)"""
        r = min(self.a, self.b, self.c)
        return tuple(list(self.center) + [r])

    # ----------------------------------------------------------------------- #
    # Sphere Approximation of Ellipsoid                                       #
    # ----------------------------------------------------------------------- #
    def approximate(self, x1=None):
        """Approximate Ellipsoid with Spheres

        This function approximates the ellipsoid by a set of spheres.
        It does so by approximating the x-z and y-z elliptical cross sections
        with circles, then scaling those circles and promoting them to spheres.

        See the documentation for
        :meth:`microstructpy.geometry.Ellipse.approximate` for more details.

        Args:
            x1 (float): *(optional)* Center position of the first sphere.
                Default is 0.75x the minimum semi-axis.

        Returns:
            numpy.ndarray: An Nx4 list of the (x, y, z, r) data of the spheres
            that approximate the ellipsoid.
        """
        if (self.a == self.b) and (self.a == self.c):
            return np.array([self.bound_max])

        if x1 is None:
            x1 = 0.25 * min(self.axes)

        # Perform approximation such that a > b > c
        if (self.a >= self.b) and (self.b >= self.c):
            a = self.a
            b = self.b
            c = self.c
            inds = [0, 1, 2]
        elif (self.a >= self.c) and (self.c >= self.b):
            a = self.a
            b = self.c
            c = self.b
            inds = [0, 2, 1]
        elif (self.b >= self.a) and (self.a >= self.c):
            a = self.b
            b = self.a
            c = self.c
            inds = [1, 0, 2]
        elif (self.c >= self.a) and (self.a >= self.b):
            a = self.c
            b = self.a
            c = self.b
            inds = [0, 2, 1]
        else:
            a = self.c
            b = self.b
            c = self.a
            inds = [2, 1, 0]

        # Prolate Ellipsoid
        if np.isclose(b, c):
            circs_xz = Ellipse(a=a, b=c).approximate()
            spheres = np.insert(circs_xz, 1, 0, axis=1)[:, np.append(inds, 3)]
            cens = np.array(self.center) + spheres[:, :-1].dot(self.matrix.T)
            spheres[:, :-1] = cens
            return spheres

        # Points in First Octant
        surface_pts = _ellipse_pts(a, b, c)
        plane_mask = np.isclose(surface_pts[:, -1], 0)

        # Circle centers
        circs_xz = Ellipse(a=a, b=c).approximate(x1)
        x_grid = np.append(circs_xz[circs_xz[:, 0] >= 0, 0][:-1], a)

        circs_yz = Ellipse(a=b, b=c).approximate(x1)
        y_grid = np.append(circs_yz[circs_yz[:, 0] >= 0, 0][:-1], b)

        xx, yy = np.meshgrid(x_grid / a, y_grid / b)

        # Conformal mapping
        xx_mesh = a * xx * np.sqrt(1 - 0.5 * yy * yy)
        yy_mesh = b * yy * np.sqrt(1 - 0.5 * xx * xx)

        xx_inner = xx_mesh[:-1, :-1].flatten()
        yy_inner = yy_mesh[:-1, :-1].flatten()
        pts_inner = np.array([xx_inner, yy_inner, 0 * xx_inner]).T

        # Interior spheres
        inner_dists = scipy.spatial.distance.cdist(pts_inner, surface_pts)

        nearest_ind = np.argmin(inner_dists, axis=1)
        nearest_dist = inner_dists[np.arange(len(nearest_ind)), nearest_ind]
        in_plane = plane_mask[nearest_ind]

        cens_inner = pts_inner[~in_plane]
        rads_inner = nearest_dist[~in_plane]

        # Exterior spheres
        xx_outer = np.concatenate((xx_mesh[:, -1], xx_mesh[-1, :-1]))
        yy_outer = np.concatenate((yy_mesh[:, -1], yy_mesh[-1, :-1]))

        u_outer = np.arctan2(yy_outer / b, xx_outer / a)
        n_x = b * np.cos(u_outer)
        n_y = a * np.sin(u_outer)
        norm_vec = np.array([n_x, n_y]).T
        norm_vec /= np.linalg.norm(norm_vec, axis=1).reshape(-1, 1)
        rads_outer = c * c * np.sqrt(n_x * n_x + n_y * n_y) / (a * b)

        rel_pos_outer = rads_outer.reshape(-1, 1) * norm_vec
        cens_outer = np.array([xx_outer, yy_outer]).T - rel_pos_outer
        cens_outer = np.hstack((cens_outer, 0 * xx_outer.reshape(-1, 1)))

        # First octant spheres
        cens = np.vstack((cens_inner, cens_outer))
        rads = np.concatenate((rads_inner, rads_outer))
        spheres = np.hstack((cens, rads.reshape(-1, 1)))

        # All octants
        for dim in range(3):
            refl_spheres = np.copy(spheres)
            mask = spheres[:, dim] > 0
            refl_spheres[mask, dim] *= -1
            spheres = np.vstack((spheres, refl_spheres[mask]))
        spheres = spheres[:, np.append(inds, 3)]
        cens = np.array(self.center) + spheres[:, :-1].dot(self.matrix.T)
        spheres[:, :-1] = cens
        return spheres

    # ----------------------------------------------------------------------- #
    # Plot Function                                                           #
    # ----------------------------------------------------------------------- #
    def plot(self, **kwargs):
        """Plot the ellipsoid.

        This function uses the :meth:`mpl_toolkits.mplot3d.Axes3D.plot_surface`
        method to add an ellipsoid to the current axes. The keyword arguments
        are passes through to the plot_surface function.

        Args:
            **kwargs (dict): Keyword arguments for matplotlib.

        """  # NOQA: E501
        if len(plt.gcf().axes) == 0:
            ax = plt.axes(projection=Axes3D.name)
        else:
            ax = plt.gca()

        u = np.linspace(0, 2 * np.pi, 11)
        cv = np.linspace(-1, 1, 12)
        uu, cvv = np.meshgrid(u, cv)
        svv = np.sin(np.arccos(cvv))
        grid_shape = uu.shape

        a = self.a
        b = self.b
        c = self.c

        xp = a * np.cos(uu) * svv
        yp = b * np.sin(uu) * svv
        zp = c * cvv

        pts = np.array([xp.reshape(1, -1).flatten(),
                        yp.reshape(1, -1).flatten(),
                        zp.reshape(1, -1).flatten()])
        rot_pts = self.orientation.dot(pts)

        xr = rot_pts[0].reshape(grid_shape)
        yr = rot_pts[1].reshape(grid_shape)
        zr = rot_pts[2].reshape(grid_shape)

        xc, yc, zc = self.center
        xx = xc + xr
        yy = yc + yr
        zz = zc + zr

        mod_kwargs = {}
        for key, val in kwargs.items():
            if key == 'facecolors' and type(val) != list:
                mod_kwargs['color'] = val
            else:
                mod_kwargs[key] = val
        ax.plot_surface(xx, yy, zz, **mod_kwargs)

    # ----------------------------------------------------------------------- #
    # Limits                                                                  #
    # ----------------------------------------------------------------------- #
    @property
    def limits(self):
        """list: List of (lower, upper) bounds for the bounding box"""
        if np.all(np.isclose(self.matrix, np.eye(3))):
            ax = np.array(self.axes)
            cen = np.array(self.center)
            return [(x - r, x + r) for x, r in zip(cen, ax)]

        n = 4
        u = np.linspace(0, 2 * np.pi, 1 + 4 * n)
        cv = np.linspace(-1, 1, 1 + 2 * n)
        uu, cvv = np.meshgrid(u, cv)
        svv = np.sin(np.arccos(cvv))

        xp = self.a * np.cos(uu) * svv
        yp = self.b * np.sin(uu) * svv
        zp = self.c * cvv

        pts = np.array([xp.flatten(), yp.flatten(), zp.flatten()])
        r_pts = self.matrix.dot(pts)
        lbs = r_pts.min(axis=-1) + np.array(self.center)
        ubs = r_pts.max(axis=-1) + np.array(self.center)
        return list(zip(lbs, ubs))

    @property
    def sample_limits(self):
        """list: List of (lower, upper) bounds for the sampling region"""
        return self.limits

    # ----------------------------------------------------------------------- #
    # Within Test                                                             #
    # ----------------------------------------------------------------------- #
    def within(self, points):
        """Test if points are within ellipsoid.

        This function tests whether a point or set of points are within the
        ellipsoid. For the set of points, a list of booleans is returned to
        indicate which points are within the ellipsoid.

        Args:
            points (list or numpy.ndarray): Point or list of points.

        Returns:
            bool or numpy.ndarray: Set to True for points in geometry.

        """
        pts = np.array(points)
        single_pt = pts.ndim == 1
        if single_pt:
            pts = pts.reshape(1, -1)

        rel_pos = pts - np.array(self.center)
        rot_pos = rel_pos.dot(self.orientation)
        scl_pos = rot_pos / np.array(self.axes).reshape(1, -1)

        sq_dist = np.sum(scl_pos * scl_pos, axis=-1)

        mask = sq_dist <= 1
        if single_pt:
            return mask[0]
        else:
            return mask

    # ----------------------------------------------------------------------- #
    # Reflect                                                                 #
    # ----------------------------------------------------------------------- #
    def reflect(self, points):
        """Reflect points across surface.

        This function reflects a point or set of points across the surface
        of the ellipsoid. Points at the center of the ellipsoid are not
        reflected.

        Args:
            points (list or numpy.ndarray): Points to reflect.

        Returns:
            numpy.ndarray: Reflected points.

        """
        pts = np.array(points)
        single_pt = pts.ndim == 1
        if single_pt:
            pts = pts.reshape(1, -1)

        rel_pos = pts - np.array(self.center)
        rot_pos = rel_pos.dot(self.orientation)
        scl_pos = rot_pos / np.array(self.axes).reshape(1, -1)

        dist = np.sqrt(np.sum(scl_pos * scl_pos, axis=-1))
        mask = dist > 0
        if not np.any(mask):
            return np.array([])

        new_dist = 2 - dist[mask]
        scl = new_dist / dist[mask]

        new_scl_pos = scl_pos[mask] * scl
        new_rel_pos = new_scl_pos.dot(self.orientation.T)
        new_pos = new_rel_pos + np.array(self.center)

        if single_pt:
            return new_pos[0]
        else:
            return new_pos


def _ellipse_arc(a, b, n):
    horiz_x = a * np.linspace(1, -1, n)
    verti_y = np.linspace(0, 2 * b, n)

    t_denom = (a - horiz_x) * verti_y + 4 * a * b
    t_numer = 4 * a * b
    t = t_numer / t_denom

    x_arc = (2 * t - 1) * a
    y_arc = t * verti_y

    return np.array([x_arc, y_arc]).T


def _ellipse_pts(a, b, c, n=81):
    arc_xy = _ellipse_arc(a, b, n)
    arc_xz = _ellipse_arc(a, c, n)
    arc_yz = _ellipse_arc(b, c, n)

    ii_xy, ii_z = np.meshgrid(np.arange(n - 1), np.arange(n - 1))
    theta_xy = np.arctan2(arc_xy[:, 1] / b, arc_xy[:, 0] / a)
    tt_xy = theta_xy[ii_xy]

    lat_xz = np.arctan2(arc_xz[:, 1] / c, arc_xz[:, 0] / a)
    lat_yz = np.arctan2(arc_yz[:, 1] / c, arc_yz[:, 0] / b)

    ll_f_xz = lat_xz[ii_z] * np.cos(tt_xy)
    ll_f_yz = lat_yz[ii_z] * np.sin(tt_xy)
    latlat = np.sqrt(ll_f_xz * ll_f_xz + ll_f_yz * ll_f_yz)

    xx = arc_xy[:, 0][ii_xy] * np.cos(latlat)
    yy = arc_xy[:, 1][ii_xy] * np.cos(latlat)
    zz = c * np.sin(latlat)

    pts = np.array([xx.flatten(), yy.flatten(), zz.flatten()]).T
    pts = np.append(pts, [[0, 0, c]], axis=0)
    return pts



"""Box

This module contains the Box class.

"""
# --------------------------------------------------------------------------- #
#                                                                             #
# Import Modules                                                              #
#                                                                             #
# --------------------------------------------------------------------------- #

from __future__ import division

import numpy as np
from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d.art3d import Poly3DCollection

from microstructpy.geometry.n_box import NBox

__author__ = 'Kenneth (Kip) Hart'


# --------------------------------------------------------------------------- #
#                                                                             #
# Box Class                                                                   #
#                                                                             #
# --------------------------------------------------------------------------- #
class Box(NBox):
    """Box

    This class contains a generic, 3D box. The position and dimensions of the
    box can be specified using any of the parameters below.

    Without any parameters, this is a unit cube centered on the origin.

    Args:
        side_lengths (list): *(optional)* Side lengths.
        center (list): *(optional)* Center of box.
        corner (list): *(optional)* Bottom-left corner.
        limits (list): *(optional)* Bounds of box.
        bounds (list): *(optional)* Alias for ``limits``.

    """
    def __init__(self, **kwargs):
        NBox.__init__(self, **kwargs)

        try:
            self.center
        except AttributeError:
            self.center = [0, 0, 0]

        try:
            self.side_lengths
        except AttributeError:
            self.side_lengths = [1, 1, 1]

    # ----------------------------------------------------------------------- #
    # Representation Function                                                 #
    # ----------------------------------------------------------------------- #
    def __repr__(self):
        repr_str = 'Box('
        repr_str += 'center=' + repr(tuple(self.center)) + ', '
        repr_str += 'side_lengths=' + repr(tuple(self.side_lengths)) + ')'
        return repr_str

    # ----------------------------------------------------------------------- #
    # Number of Dimensions                                                    #
    # ----------------------------------------------------------------------- #
    @property
    def n_dim(self):
        """int: number of dimensions, 3"""
        return 3

    # ----------------------------------------------------------------------- #
    # Volume                                                                  #
    # ----------------------------------------------------------------------- #
    @property
    def volume(self):
        """float: volume of box, :math:`V=l_1 l_2 l_3`"""
        return self.n_vol

    # ----------------------------------------------------------------------- #
    # Plot Function                                                           #
    # ----------------------------------------------------------------------- #
    def plot(self, **kwargs):
        """Plot the box.

        This function adds an
        :class:`mpl_toolkits.mplot3d.art3d.Poly3DCollection` to the current
        axes. The keyword arguments are passed through to the Poly3DCollection.

        Args:
            **kwargs (dict): Keyword arguments for Poly3DCollection.

        """  # NOQA: E501
        if len(plt.gcf().axes) == 0:
            ax = plt.axes(projection=Axes3D.name)
        else:
            ax = plt.gca()

        xlim, ylim, zlim = self.limits

        inds = [(0, 0), (0, 1), (1, 1), (1, 0)]

        # x faces
        f1 = np.array([(xlim[0], ylim[i], zlim[j]) for i, j in inds])
        f2 = np.array([(xlim[1], ylim[i], zlim[j]) for i, j in inds])

        # y faces
        f3 = np.array([(xlim[i], ylim[0], zlim[j]) for i, j in inds])
        f4 = np.array([(xlim[i], ylim[1], zlim[j]) for i, j in inds])

        # z faces
        f5 = np.array([(xlim[i], ylim[j], zlim[0]) for i, j in inds])
        f6 = np.array([(xlim[i], ylim[j], zlim[1]) for i, j in inds])

        # plot
        xy = [f1, f2, f3, f4, f5, f6]
        pc = Poly3DCollection(xy, **kwargs)
        ax.add_collection(pc)


# --------------------------------------------------------------------------- #
#                                                                             #
# Cube Class                                                                  #
#                                                                             #
# --------------------------------------------------------------------------- #
class Cube(Box):
    """A cube.

    This class contains a generic, 3D cube. It is derived from the
    :class:`.Box` and contains the ``side_length`` property, rather than
    multiple side lengths.

    Without any parameters, this is a unit cube centered on the origin.

    Args:
        side_length (float): *(optional)* Side length.
        center (list, tuple, numpy.ndarray): *(optional)* Center of box.
        corner (list, tuple, numpy.ndarray): *(optional)* Bottom-left corner.
    """
    def __init__(self, **kwargs):
        if 'side_length' in kwargs:
            kwargs['side_lengths'] = 3 * [kwargs['side_length']]

        Box.__init__(self, **kwargs)

    # ----------------------------------------------------------------------- #
    # Side Length Property                                                    #
    # ----------------------------------------------------------------------- #
    @property
    def side_length(self):
        """float: length of the side of the cube."""
        return self.side_lengths[0]



# --------------------------------------------------------------------------- #
#                                                                             #
# Import Modules                                                              #
#                                                                             #
# --------------------------------------------------------------------------- #
from __future__ import division

import numpy as np

__author__ = 'Kenneth (Kip) Hart'


# --------------------------------------------------------------------------- #
#                                                                             #
# NSphere Class                                                               #
#                                                                             #
# --------------------------------------------------------------------------- #
class NSphere(object):
    """An N-dimensional sphere.

    This class represents a generic, n-dimensional sphere. It is defined by
    a center point and size parameter, which can be either radius or diameter.

    If multiple size or position keywords are given, there is no guarantee
    whhich keywords are used to create the geometry.

    Args:
        r (float): *(optional)* The radius of the n-sphere.
            Defaults to 1.
        center (list): *(optional)* The coordinates of the center.
            Defaults to [].
        radius : Alias for ``r``.
        d : Alias for ``2*r```.
        diameter : Alias for ``2*r``.
        size : Alias for ``2*r``.
        position : Alias for ``center``.

    """
    # ----------------------------------------------------------------------- #
    # Constructor                                                             #
    # ----------------------------------------------------------------------- #
    def __init__(self, **kwargs):
        if 'r' in kwargs:
            self.r = kwargs['r']
        elif 'radius' in kwargs:
            self.r = kwargs['radius']
        elif 'd' in kwargs:
            self.r = 0.5 * kwargs['d']
        elif 'diameter' in kwargs:
            self.r = 0.5 * kwargs['diameter']
        elif 'size' in kwargs:
            self.r = 0.5 * kwargs['size']
        else:
            self.r = 1

        if 'center' in kwargs:
            self.center = kwargs['center']
        elif 'position' in kwargs:
            self.center = kwargs['position']
        else:
            self.center = []

    @classmethod
    def best_fit(cls, points):
        """Find n-sphere of best fit for set of points.

        This function takes a list of points and computes an n-sphere of
        best fit, in an algebraic sense. This method was developed using the
        a published writeup, which was extended from 2D to ND. [#bullock]_

        Args:
            points (list, numpy.ndarray): List of points to fit.

        Returns:
            NSphere: An instance of the class that fits the points.

        .. [#bullock] Circle fitting writup by Randy Bullock,
          https://dtcenter.org/met/users/docs/write_ups/circle_fit.pdf
        """  # NOQA: E501
        # convert points to numpy array
        pts = np.array(points)
        n_pts, n_dim = pts.shape
        if n_pts <= n_dim:
            mid = pts.mean(axis=0)
            rel_pos = pts - mid
            dist = np.linalg.norm(rel_pos, axis=1).mean()
            return cls(center=mid, radius=dist)

        # translate points to average position
        bcenter = pts.mean(axis=0)
        pts -= bcenter

        # Assemble matrix and vector of sums
        mat = np.zeros((n_dim, n_dim))
        vec = np.zeros(n_dim)

        for i in range(n_dim):
            for j in range(n_dim):
                mat[i, j] = np.sum(pts[:, i] * pts[:, j])
                vec[i] += np.sum(pts[:, i] * pts[:, j] * pts[:, j])
        vec *= 0.5

        # Solve linear system for the center
        try:
            cen_b = np.linalg.solve(mat, vec)
        except np.linalg.linalg.LinAlgError:
            cen_b = pts.mean(axis=0)
        cen = cen_b + bcenter

        # Calculate the radius
        alpha = np.sum(cen_b * cen_b) + np.trace(mat) / n_pts
        R = np.sqrt(alpha)

        # Create the instance
        return cls(center=cen, radius=R)

    # ----------------------------------------------------------------------- #
    # String and Representation Functions                                     #
    # ----------------------------------------------------------------------- #
    def __str__(self):
        str_str = 'Radius: ' + str(self.r) + '\n'
        str_str += 'Center: ' + str(tuple(self.center))
        return str_str

    def __repr__(self):
        repr_str = 'NSphere('
        repr_str += 'r=' + repr(self.r) + ', '
        repr_str += 'center=' + repr(tuple(self.center)) + ')'
        return repr_str

    # ----------------------------------------------------------------------- #
    # Equality                                                                #
    # ----------------------------------------------------------------------- #
    def __eq__(self, nsphere):
        if not hasattr(nsphere, 'r'):
            return False

        if not np.isclose(self.r, nsphere.r):
            return False

        if not hasattr(nsphere, 'center'):
            return False

        c1 = np.array(self.center)
        c2 = np.array(nsphere.center)

        if c1.shape != c2.shape:
            return False

        dx = np.array(self.center) - np.array(nsphere.center)
        if not np.all(np.isclose(dx, 0)):
            return False
        return True

    def __neq__(self, nsphere):
        return not self.__eq__(nsphere)

    # ----------------------------------------------------------------------- #
    # Size Setters/Getters                                                    #
    # ----------------------------------------------------------------------- #
    @property
    def radius(self):
        """float: radius of n-sphere."""
        return self.r

    @property
    def d(self):
        """float: diameter of n-sphere."""
        return 2 * self.r

    @property
    def diameter(self):
        """float: diameter of n-sphere."""
        return 2 * self.r

    @property
    def size(self):
        """float: size (diameter) of n-sphere."""
        return 2 * self.r

    @property
    def position(self):
        """list: position of n-sphere."""
        return self.center

    # ----------------------------------------------------------------------- #
    # Bounding N-Spheres                                                      #
    # ----------------------------------------------------------------------- #
    @property
    def bound_max(self):
        """tuple: maximum bounding n-sphere"""
        return tuple(list(self.center) + [self.r])

    @property
    def bound_min(self):
        """tuple: minimum interior n-sphere"""
        return self.bound_max

    # ----------------------------------------------------------------------- #
    # Limits                                                                  #
    # ----------------------------------------------------------------------- #
    @property
    def limits(self):
        """list: list of (lower, upper) bounds for the bounding box"""
        return [(x - self.r, x + self.r) for x in self.center]

    @property
    def sample_limits(self):
        """list: list of (lower, upper) bounds for the sampling region"""
        return self.limits

    # ----------------------------------------------------------------------- #
    # Approximate                                                             #
    # ----------------------------------------------------------------------- #
    def approximate(self):
        """Approximate the n-sphere with itself

        Other geometries can be approximated by a set of circles or spheres.
        For the n-sphere, this approximation is exact.

        Returns:
            list: A list containing [(x, y, z, ..., r)]
        """
        return [tuple(list(self.center) + [self.r])]

    # ----------------------------------------------------------------------- #
    # Within Test                                                             #
    # ----------------------------------------------------------------------- #
    def within(self, points):
        """Test if points are within n-sphere.

        This function tests whether a point or set of points are within the
        n-sphere. For the set of points, a list of booleans is returned to
        indicate which points are within the n-sphere.

        Args:
            points (list or numpy.ndarray): Point or list of points.

        Returns:
            bool or numpy.ndarray: Set to True for points in geometry.
        """
        pts = np.array(points)
        single_pt = pts.ndim == 1
        if single_pt:
            pts = pts.reshape(1, -1)

        rel_pos = pts - np.array(self.center)
        sq_dist = np.sum(rel_pos * rel_pos, axis=-1)

        mask = sq_dist <= self.r * self.r
        if single_pt:
            return mask[0]
        else:
            return mask

    # ----------------------------------------------------------------------- #
    # Reflect                                                                 #
    # ----------------------------------------------------------------------- #
    def reflect(self, points):
        """Reflect points across surface.

        This function reflects a point or set of points across the surface
        of the n-sphere. Points at the center of the n-sphere are not
        reflected.

        Args:
            points (list or numpy.ndarray): Points to reflect.

        Returns:
            numpy.ndarray: Reflected points.
        """
        pts = np.array(points)
        single_pt = pts.ndim == 1
        if single_pt:
            pts = pts.reshape(1, -1)

        rel_pos = pts - np.array(self.center)
        cen_dist = np.sqrt(np.sum(rel_pos * rel_pos, axis=-1))
        mask = cen_dist > 0
        new_dist = 2 * self.r - cen_dist[mask]
        scl = new_dist / cen_dist[mask]

        new_rel_pos = np.zeros(pts.shape)
        new_rel_pos[mask] = scl.reshape(-1, 1) * rel_pos[mask]
        new_rel_pos[~mask] = 0
        new_rel_pos[~mask, 0] = 2 * self.r

        new_pts = new_rel_pos + np.array(self.center)
        if single_pt:
            return new_pts[0]
        else:
            return new_pts



"""Seed

This module contains the class definition for the Seed class.
"""

# --------------------------------------------------------------------------- #
#                                                                             #
# Import Modules                                                              #
#                                                                             #
# --------------------------------------------------------------------------- #

from __future__ import division
from __future__ import print_function

import numpy as np
from matplotlib import collections
from matplotlib import patches
from matplotlib import pyplot as plt

from microstructpy import _misc
from microstructpy import geometry

__all__ = ['Seed']
__author__ = 'Kenneth (Kip) Hart'


# --------------------------------------------------------------------------- #
#                                                                             #
# Seed Class                                                                  #
#                                                                             #
# --------------------------------------------------------------------------- #
class Seed(object):
    """Seed particle

    The Seed class contains the information about a single seed in the mesh.
    These seeds have a geometry (from :mod:`microstructpy.geometry`),
    a phase number, a breakdown, and a position.

    Args:
        seed_geometry (from :mod:`microstructpy.geometry`) : The geometry of
            the seed.
        phase (int) : *(optional)* The phase number of the seed.
            Defaults to 0.
        breakdown (list or numpy.ndarray) : *(optional)* The circle/sphere
            approximation of this grain. The format for this input is::

                #                 x   y  r
                breakdown_2D = [( 2,  3, 1),
                                ( 0,  0, 4),
                                (-2,  4, 8)]

                #                 x   y   z  r
                breakdown_3D = [( 3, -1,  2, 1),
                                ( 0,  2, -1, 1)]

            The default behavior is to call the ``approximate()`` function
            of the geometry.
        position (list or numpy.ndarray) : *(optional)* The coordinates of the
            seed. See :attr:`position` for more details.
            Defaults to the origin.
    """
    # ----------------------------------------------------------------------- #
    # Initializer                                                             #
    # ----------------------------------------------------------------------- #
    def __init__(self, seed_geometry, phase=0, breakdown=None, position=None):
        self.geometry = seed_geometry
        self.phase = phase

        if position is None and self.geometry is not None:
            self.position = [0 for _ in range(self.geometry.n_dim)]
        else:
            self.position = position

        if breakdown is None:
            self.breakdown = seed_geometry.approximate()
        else:
            self.breakdown = breakdown

    # ----------------------------------------------------------------------- #
    # Factory Method                                                          #
    # ----------------------------------------------------------------------- #
    @classmethod
    def factory(cls, seed_type, phase=0, breakdown=None, position=None,
                **kwargs):
        """Factory method for seeds

        This function returns a seed based on the seed type and keyword
        arguments associated with that type. The currently supported types
        are:

            * circle
            * ellipse
            * ellipsoid
            * rectangle
            * sphere
            * square

        If the seed_type is not on this list, an error is thrown.

        Args:
            seed_type (str): type of seed, from list above.
            phase (int): *(optional)* Material phase number of seed.
                Defaults to 0.
            breakdown (list or numpy.ndarray): *(optional)* List of circles or
                spheres that approximate the geometry. The list should be
                formatted as follows::

                    breakdown = [(x1, y1, z1, r1),
                                 (x2, y2, z2, r2),
                                 ...]

                The breakdown will be automatically generated if not provided.
            position (list or numpy.ndarray): *(optional)* The coordinates of
                the seed. Default is the origin.
            **kwargs: Keyword arguments that define the size, shape, etc of the
                seed geometry.

        Returns:
            Seed: An instance of the class.
        """
        assert type(seed_type) is str
        seed_type = seed_type.strip().lower()

        if seed_type == 'nonetype':
            n_dim = 0
        else:
            n_dim = geometry.factory(seed_type).n_dim
        if 'volume' in kwargs:
            if n_dim == 2:
                size = 2 * np.sqrt(kwargs['volume'] / np.pi)
            else:
                size = 2 * np.cbrt(3 * kwargs['volume'] / (4 * np.pi))
            kwargs['size'] = size

        # Catch NoneType geometries
        if seed_type == 'nonetype':
            geom = None
        else:
            geom = geometry.factory(seed_type, **kwargs)

        if breakdown is None:
            if seed_type in ('circle', 'sphere'):
                breakdown = np.append(geom.center, geom.r).reshape(1, -1)
            else:
                breakdown = geom.approximate()

        if position is None:
            position = [0 for _ in range(geom.n_dim)]

        return cls(geom, phase, breakdown, position)

    # ----------------------------------------------------------------------- #
    # Read from String                                                        #
    # ----------------------------------------------------------------------- #
    @classmethod
    def from_str(cls, seed_str):
        """Create seed from a string.

        This method creates a seed particle from a string representation.
        This is used when reading in seeds from a file.

        Args:
            seed_str (str): String representation of the seed.

        Returns:
            Seed: An instance of a Seed derived class.

        """
        # Convert to dictionary
        str_dict = {}
        for line in seed_str.strip().split('\n'):
            try:
                k_str, v_str = line.split(':')
            except ValueError:
                continue
            else:
                k = k_str.strip().lower().replace(' ', '_')
                v = _misc.from_str(v_str)
                str_dict[k] = v

        # Extract seed type, phase, and breakdown
        seed_type = str_dict['geometry']
        del str_dict['geometry']

        if 'phase' in str_dict:
            phase = str_dict['phase']
            del str_dict['phase']
        else:
            phase = 0

        if 'breakdown' in str_dict:
            breakdown = str_dict['breakdown']
            del str_dict['breakdown']
        else:
            breakdown = None

        if 'position' in str_dict:
            position = str_dict['position']
            del str_dict['position']
        else:
            position = None

        return cls.factory(seed_type, phase, breakdown, position, **str_dict)

    # ----------------------------------------------------------------------- #
    # String and Representation Functions                                     #
    # ----------------------------------------------------------------------- #
    def __str__(self):
        geom_name = type(self.geometry).__name__.lower()
        str_str = 'Geometry: ' + geom_name + '\n'
        str_str += str(self.geometry) + '\n'
        str_str += 'Phase: ' + str(self.phase) + '\n'
        bkdwn_str = ', '.join([str(tuple(b)) for b in self.breakdown])
        str_str += 'Breakdown: (' + bkdwn_str + ')\n'
        str_str += 'Position: (' + ', '.join([str(x) for x in self.position])
        str_str += ')'
        return str_str

    def __repr__(self):
        repr_str = 'Seed('
        repr_str += repr(self.geometry) + ', '
        repr_str += 'phase=' + repr(self.phase) + ', '
        bkdwn_str = ', '.join([repr(tuple(b)) for b in self.breakdown])
        repr_str += 'breakdown=(' + bkdwn_str + '), '
        repr_str += 'position=(' + ', '.join([repr(x) for x in self.position])
        repr_str += ')'
        repr_str += ')'
        return repr_str

    # ----------------------------------------------------------------------- #
    # Comparison Functions                                                    #
    # ----------------------------------------------------------------------- #
    def __lt__(self, seed):
        if self.geometry is None:
            return True
        if seed.geometry is None:
            return False

        a_str = 'Seeds are not the same dimension.'
        assert self.geometry.n_dim == seed.geometry.n_dim, a_str
        if self.geometry.n_dim == 2:
            v_self = self.geometry.area
            v_seed = seed.geometry.area
        else:
            v_self = self.geometry.volume
            v_seed = seed.geometry.volume

        if v_self == v_seed:
            return self.phase < seed.phase
        else:
            return v_self < v_seed

    def __eq__(self, seed):
        if not isinstance(seed, Seed):
            return False

        if seed.phase != self.phase:
            return False

        if not np.all(np.isclose(seed.breakdown, self.breakdown)):
            return False

        if seed.geometry != self.geometry:
            return False

        if not np.all(np.isclose(seed.position, self.position)):
            return False

        return True

    # ----------------------------------------------------------------------- #
    # Position Getter/Setter                                                  #
    # ----------------------------------------------------------------------- #
    @property
    def position(self):
        """Position of the seed

        This is the location of the seed center.

        Note:
            If the breakdown of the seed has been populated, the setter
            function will update the position of the center and translate
            the breakdown circles/spheres.

        """
        return self._position

    @position.setter
    def position(self, pos):
        try:
            old_pos = np.array(self.position)
        except AttributeError:
            old_pos = np.zeros(len(pos))

        try:
            displace = np.array(pos) - old_pos
            for i, bkdwn in enumerate(self.breakdown):
                coords = bkdwn[:-1]
                rad = bkdwn[-1]
                new_coords = [x + d for x, d in zip(coords, displace)]
                new_bkdwn = new_coords + [rad]
                self.breakdown[i] = new_bkdwn
        except AttributeError:
            pass

        try:
            self.geometry.center = pos
        except AttributeError:
            pass

        self._position = pos

    # ----------------------------------------------------------------------- #
    # Volume                                                                  #
    # ----------------------------------------------------------------------- #
    @property
    def volume(self):
        """float: The area (2D) or volume (3D) of the seed"""
        if self.geometry is None:
            return 0
        if self.geometry.n_dim == 2:
            return self.geometry.area
        else:
            return self.geometry.volume

    # ----------------------------------------------------------------------- #
    # Limits                                                                  #
    # ----------------------------------------------------------------------- #
    @property
    def limits(self):
        """list: The (lower, upper) bounds of the seed"""
        if self.geometry is None:
            return []
        return self.geometry.limits

    # ----------------------------------------------------------------------- #
    # Plot                                                                    #
    # ----------------------------------------------------------------------- #
    def plot(self, **kwargs):
        """Plot the seed

        This function plots the geometry of the seed. The keyword arguments
        are passed through to matplotlib.
        See the plot methods in :mod:`microstructpy.geometry` for more
        details.

        Args:
            **kwargs: Plotting keyword arguments.

        """
        if self.geometry is not None:
            self.geometry.plot(**kwargs)

    # ----------------------------------------------------------------------- #
    # Plot Breakdown                                                          #
    # ----------------------------------------------------------------------- #
    def plot_breakdown(self, **kwargs):
        """Plot breakdown of seed

        This function plots the circle/sphere breakdown of the seed. In 2D,
        this adds a :class:`matplotlib.collections.PatchCollection`
        to the current axes.

        Args:
            **kwargs: Matplotlib keyword arguments.

        """

        n = len(self.breakdown[0]) - 1
        if n == 2:
            pc = [patches.Circle([x, y], r) for x, y, r in self.breakdown]
            coll = collections.PatchCollection(pc, **kwargs)
            plt.gca().add_collection(coll)

        else:
            [geometry.Sphere(r=r, center=(x, y, z)).plot(**kwargs)
             for x, y, z, r in self.breakdown]



from __future__ import division

import os

import numpy as np
import scipy.stats
from matplotlib import pyplot as plt

import microstructpy as msp


def main():
    # Colors
    c1 = '#12C2E9'
    c2 = '#C471ED'
    c3 = '#F64F59'

    # Offset
    off = 1

    # Create Directory
    dirname = os.path.join(os.path.dirname(__file__), 'docs_banner')
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    # Create Domain
    domain = msp.geometry.Rectangle(width=10, length=20)

    # Create Unpositioned Seeds
    phase2 = {'color': c1}
    ell_geom = msp.geometry.Ellipse(a=8, b=3)
    ell_seed = msp.seeding.Seed(ell_geom, phase=2)

    mu = 1
    bnd = 0.5
    d_dist = scipy.stats.uniform(loc=mu-bnd, scale=2*bnd)
    phase0 = {'color': c2, 'shape': 'circle', 'd': d_dist}
    phase1 = {'color': c3, 'shape': 'circle', 'd': d_dist}
    circle_area = domain.area - ell_geom.area
    seeds = msp.seeding.SeedList.from_info([phase0, phase1], circle_area)

    seeds.append(ell_seed)
    hold = [False for seed in seeds]
    hold[-1] = True
    phases = [phase0, phase1, phase2]

    # Create Positioned Seeds
    seeds.position(domain, hold=hold, verbose=True)

    # Create Polygonal Mesh
    pmesh = msp.meshing.PolyMesh.from_seeds(seeds, domain)

    # Create Triangular Mesh
    tmesh = msp.meshing.TriMesh.from_polymesh(pmesh,
                                              min_angle=12,
                                              max_edge_length=0.2,
                                              max_volume=0.4)

    # Create Figure
    k = 0.12
    len_x = 3 * domain.length + 4 * off
    len_y = domain.width + 2 * off
    plt.figure(figsize=(k * len_x, k * len_y))

    # Plot Seeds
    seed_colors = [phases[s.phase]['color'] for s in seeds]
    seeds.plot(color=seed_colors, alpha=0.8, edgecolor='k', linewidth=0.3)
    domain.plot(facecolor='none', edgecolor='k', linewidth=0.3)

    # Plot Polygonal Mesh
    pmesh.points = np.array(pmesh.points)
    pmesh.points[:, 0] += domain.length + off
    for region, phase_num in zip(pmesh.regions, pmesh.phase_numbers):
        if phase_num == 2:
            continue
        color = phases[phase_num]['color']

        facets = [pmesh.facets[f] for f in region]
        kps = ordered_kps(facets)
        x, y = zip(*[pmesh.points[kp] for kp in kps])
        plt.fill(x, y, color=color, alpha=0.8, edgecolor='none')

    ellipse_regions = set()
    for region_num, phase_num in enumerate(pmesh.phase_numbers):
        if phase_num == 2:
            ellipse_regions.add(region_num)

    ellipse_facets = []
    for facet, neighbors in zip(pmesh.facets, pmesh.facet_neighbors):
        common_regions = ellipse_regions & set(neighbors)
        if len(common_regions) == 1:
            ellipse_facets.append(facet)
    ellipse_kps = ordered_kps(ellipse_facets)
    x, y = zip(*[pmesh.points[kp] for kp in ellipse_kps])
    plt.fill(x, y, color=phases[-1]['color'], alpha=0.8, edgecolor='none')

    for facet, neighbors in zip(pmesh.facets, pmesh.facet_neighbors):
        common_regions = ellipse_regions & set(neighbors)
        if len(common_regions) < 2:
            x, y = zip(*[pmesh.points[kp] for kp in facet])
            plt.plot(x, y, color='k', linewidth=0.3)

    # Plot Triangular Mesh
    tmesh.points = np.array(tmesh.points)
    tmesh.points[:, 0] += 2 * off + 2 * domain.length
    tri_colors = [seed_colors[n] for n in tmesh.element_attributes]
    tmesh.plot(color=tri_colors, alpha=0.8, edgecolor='k', linewidth=0.2)

    # Set Up Axes
    plt.gca().set_position([0, 0, 1, 1])
    plt.axis('image')
    plt.gca().set_axis_off()
    plt.gca().get_xaxis().set_visible(False)
    plt.gca().get_yaxis().set_visible(False)

    xlim, ylim = domain.limits
    xlim[0] -= off
    xlim[1] += 3 * off + 2 * domain.length

    ylim[0] -= off
    ylim[1] += off

    plt.axis(list(xlim) + list(ylim))

    fname = os.path.join(dirname, 'banner.png')
    plt.savefig(fname, bbox='tight', pad_inches=0)
    plt.savefig(fname.replace('.png', '.pdf'), bbox='tight', pad_inches=0)


def ordered_kps(pairs):
    t_pairs = [tuple(p) for p in pairs]
    kps = list(t_pairs.pop())
    while t_pairs:
        for i, pair in enumerate(t_pairs):
            if kps[-1] in pair:
                break
        assert kps[-1] in pair, pairs
        kps += [kp for kp in t_pairs.pop(i) if kp != kps[-1]]
    return kps[:-1]


if __name__ == '__main__':
    main()
